#include <stdio.h>
#include <math.h>
#include <stdlib.h>

/*Note that blas_dgemm declared in gsl_blas.h doesn't allow 
  COLUMN-major matrix storage option*/
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

void MN_sampler(double *v_MH_mean, double *MKL_Chol_Cov, 
				double scale,  int N, gsl_rng *r, double *mult_norm);

void MT_sampler(double *v_MH_mean, double *MKL_Chol_Cov, double scale,
				double MT_mu, int N, gsl_rng *r, double *mult_t);

double MT_kernel(double * v_MH_mean, double * MKL_T_InvCov,
				 double scale, double MT_mu, int N, double *THETA_MH_i);

/************************************************************************/
/* THIS FUNCTION MAKES A DRAW FROM N-DIMENSIONAL NORMAL DISTRIBUTION    */
/*                                                                      */ 
/* OUTPUT: N-dimensional array from Multivariate Normal                 */
/*                                                                      */
/* INPUT:  mean and L - Low-Triangular Cholesky(Covariance) arrays in   */
/*         column-major (Fortran-like) form, scalor used to scale L     */
/*         Cholesky(Covar) matrix (in particular useful for student-t   */
/*         generation), # dimensions, random number generator           */
/*                                                                      */
/* NOTE:   NOT Covariance, but L from SIGMA = L*L^T Cholesky decomp     */
/*         is used as an argument. Again, because it is costly to       */
/*         compute Cholesky Decomp all the time inside M-H loop.        */
/*         ALSO NOTE THAT BECAUSE VAR(A*X)=A*VAR(X)*A', THE INPUT is    */
/*         L, NOT L^T!!!                                                */
/************************************************************************/

void MN_sampler(double *v_MH_mean, double *MKL_Chol_Cov, 
				double scale,  int N, gsl_rng *_gsl_r, double *mult_norm)
{
	int i;
	double *iid_norm = new double [N];

	for (i=0;i<N;i++)
		{
		iid_norm[i] = gsl_ran_gaussian(_gsl_r, 1.0);
		mult_norm[i] = v_MH_mean[i];
		}

	/******mult_norm = scale*MKL_Chol_Cov * iid_norm + v_MH_mean****/
	cblas_dgemm (CblasColMajor, CblasNoTrans, CblasNoTrans, N, 1, N, 
				scale, MKL_Chol_Cov, N, iid_norm, N, 1.0, mult_norm, N);

	delete [] iid_norm;
}



/************************************************************************/
/* THIS FUNCTION MAKES A DRAW FROM N-DIMENSIONAL STUDENT-T DISTRIBUTION */
/* BY FIRST DRAWING lambda FROM GAMMA(nu/2,nu/2), AND THEN DRAWING FROM */
/* MULTIVARIATE NORMAL WITH MEAN mu AND COVARIANCE (lambda)^-1 * SIGMA  */
/*                                                                      */ 
/* OUTPUT: N-dimensional array from Multivariate Student-T              */
/*                                                                      */
/* INPUT:  mean and L - Low-Triangular Cholesky(Covariance) arrays in   */
/*         column-major (Fortran-like) form, scalor used to scale L     */
/*         Cholesky(Covar) matrix (in particular useful if we want to   */
/*         implement SCALED Random Walk M-H), degrees of freedom,       */
/*         # dimensions, random number generator                        */
/*                                                                      */
/* NOTE:   NOT Covariance, but L from SIGMA = L*L^T Cholesky decomp     */
/*         is used as an argument. Again, because it is costly to       */
/*         compute Cholesky Decomp all the time inside M-H loop.        */
/*         ALSO NOTE THAT BECAUSE VAR(A*X)=A*VAR(X)*A', THE INPUT is    */
/*         L, NOT L^T!!!                                                */
/* NOTE:   IF YOUR OBJECTIVE IS TO GET A DRAW FROM MULT. STUDENT-T      */
/*         WITH VARIANCE EXACTLY EQUAL TO NEGATIVE INVERSE HESSIAN,     */
/*         ONE MUST REMEMBER TO SET scale = sqrt((MT_mu-2)/MT_mu)       */
/*         BECAUSE FOR X~T(NU,SIGMA,MU), VAR(X)=SIGMA*MU/(MU-2)         */
/************************************************************************/

void MT_sampler(double *v_MH_mean, double *MKL_Chol_Cov, double scale,
				double MT_mu, int N, gsl_rng *_gsl_r, double *mult_t)
{
	double lambda;
	lambda = gsl_ran_gamma(_gsl_r, MT_mu/2.0, 2.0/MT_mu);

	MN_sampler(v_MH_mean, MKL_Chol_Cov, scale*pow(lambda,-0.5), 
			   N, _gsl_r, mult_t);

}



/******************************************************************/
/* THIS FUNCTION CALCULATES KERNEL (PDF W/O NORMALIZING CONSTANT) */
/* OF N-DIMENSIONAL STUDENT-T DISTRIBUTION                        */
/*                                                                */
/* OUTPUT: (double) pdf value                                     */
/*                                                                */
/* INPUT: mean and INV(variance) arrays in column-major form,     */
/*        scalor for Covar, degrees of freedom, # dimensions,     */
/*        array N-dim point at which kernel is calculated         */
/*                                                                */
/* NOTE: INVERSE(!!!) OF THE COVARIANCE MATRIX IS USED HERE!!     */
/*       THAT IS BECAUSE INVERSE IS USUALLY AVAILABLE IN TAILORED */
/*       APPROACH EVEN BEFORE THE VARIANCE IS AVAILABLE.          */
/*       SO WE CAN AVOID UNNECESSARY INVERSE COMPUTATION          */
/*       WHICH IS TOO COSTLY INSIDE M-H LOOP                      */
/******************************************************************/


double MT_kernel(double * v_MH_mean, double * MKL_T_InvCov,
				 double scale, double MT_mu, int N, double *THETA_MH_i)
{
	int i;
	double *demeaned = new double [N];
	double *temp1 = new double [N];
	double kernelm[1];
	double ans;

	for (i=0;i<N;i++)
		demeaned[i] = THETA_MH_i[i] - v_MH_mean[i];

	//******temp1 = (demeaned)' * MKL_T_InvCov
	cblas_dgemm (CblasColMajor, CblasTrans, CblasNoTrans, 1, N, N,  
				scale, demeaned, N, MKL_T_InvCov, N, 0.0, temp1, 1);

	//******kernelm = temp1 * demeaned
	cblas_dgemm (CblasColMajor, CblasNoTrans, CblasNoTrans, 1, 
				1, N, 1.0, temp1, 1, demeaned, N, 0.0, kernelm, 1);

	ans = pow((1.0+kernelm[0]/MT_mu),(-(MT_mu+(double) N)/2.0));

	delete [] demeaned;
	delete [] temp1;

	return ans;
}

