#include<cmath>
#include<limits>
#include<gsl/gsl_matrix.h>
#include<gsl/gsl_errno.h>
#include"dfridders.h"

extern "C" {

  class Dfridders {
  public:
    Dfridders( size_t n = 10, double stepDec = 1.4 )
      : n( n ), stepDec( stepDec ), stepDec2( stepDec * stepDec ){  
      tableau = gsl_matrix_alloc( n, n );
    }
    ~Dfridders(){
      if( tmp != 0 )
	gsl_vector_free( tmp );
      gsl_matrix_free( tableau );
    }
    int operator()( gsl_vector* g, gsl_vector* error,
		    double (*function)( const gsl_vector*, void* ),
		    const gsl_vector* x, void* params,
		    const double h ) const {
      /* check range of h */
      if( h < 0 )
	return GSL_EDOM;
      /* sanity checks */
      if( x == 0 || function == 0 )
	return GSL_EINVAL;
      if( error != 0 && error->size != x->size )
	return GSL_EINVAL;
    
      /* compute */
      double err = 0;
      /* allocate tmp if need be */
      if( tmp == 0 || tmp->size != x->size ){
	if( tmp != 0 )
	  gsl_vector_free( tmp );
	tmp = gsl_vector_alloc( x->size );
      }
      /* copy to tmp */
      gsl_vector_memcpy( tmp, x );
      for( size_t i = 0; i < x->size; ++i ){
	/* define f from function */
	double xi = df( &err, function, params, h, i );
	gsl_vector_set( g, i, xi );
	gsl_vector_set( error, i, err );
      }
    }
  
  private:
    /* this function actually does the work */
    double df( double* err, double (*function)( const gsl_vector*, void* ),
	       void* params, const double h, const size_t i ) const {
      /* note that x is copied to tmp */
      double hc = h;
      const double xc = gsl_vector_get( tmp, i );
      *err = std::numeric_limits<double>::max();

      /* initialise */
      gsl_vector_set( tmp, i, xc + hc );
      double fp = function( tmp, params );
      gsl_vector_set( tmp, i, xc - hc );
      double fm = function( tmp, params );
      gsl_matrix_set( tableau, 0, 0, (fp - fm) / (2 * hc) );

      double result = 0;
    
      /* main loop */
      for( size_t j = 1; j < n; ++j ){
	hc *= scale;
	gsl_vector_set( tmp, i, xc + hc );
	double fp = function( tmp, params );
	gsl_vector_set( tmp, i, xc - hc );
	double fm = function( tmp, params );
	gsl_matrix_set( tableau, 0, j, (fp - fm) / (2 * hc) );
      
	/* second loop */
	double d = scale2;
	for( size_t k = 0; k < j; ++k ){
	  fp = gsl_matrix_get( tableau, k, j ) * d;
	  fm = gsl_matrix_get( tableau, k, j - 1 );
	  gsl_matrix_set( tableau, k + 1 , j, (fp - fm) / (d - 1) );
	  d *= scale2;
	  double error = (fp = std::fabs( gsl_matrix_get( tableau, k + 1, j )
					  - gsl_matrix_get( tableau, k, j ) ) )
	    > (fm = std::fabs( gsl_matrix_get( tableau, k + 1, j )
			       - gsl_matrix_get( tableau, k, j - 1 ) ) ) ? fp : fm;
	  if( error <= *err ){
	    *err = error;
	    result = gsl_matrix_get( tableau, k + 1, j );
	  }
	}
	/* check that new result is not substantially worse */
	if( std::fabs( gsl_matrix_get( tableau, j, j )
		       - gsl_matrix_get( tableau, j - 1, j - 1 )) >= 2 * *err )
	  break;
      }
    
      /* tidy up */
      gsl_vector_set( tmp, i, xc );

      return result;
    }

  private:
    mutable gsl_matrix* tableau;
    const size_t n;
    const double stepDec;
    const double stepDec2;
    mutable gsl_vector* tmp;
    static const double scale;
    static const double scale2;
  } df0, df1, df2;

  const double Dfridders::scale = static_cast<double>( 1 ) / 1.4;
  const double Dfridders::scale2 = 1.4 * 1.4;

  int dfridders( const Dfridders* df,
		 gsl_vector* g,
		 gsl_vector* error,
		 double (*function)( const gsl_vector*, void* ),
		 const gsl_vector* x,
		 void* params,
		 const double h ){
    return (*df)( g, error, function, x, params, h );
  }
  
  Dfridders* get_dfridders( int i ){
    switch( i ){
    case 0:
      return& df0;
      break;
    case 1:
      return& df1;
      break;
    case 2:
      return& df2;
      break;
    default:
      return 0;
    }
  }
  
}
