#include <gsl/gsl_math.h>
#include <gsl/gsl_matrix.h>
#include <math.h>
#define _con 1.4
#define _con2 (_con*_con)
#define _big 1.0e30
#define _ntab 10
#define _safe 2.0


double my_dfridr(const gsl_function *f, double x, double h, 
		 double *err);

double my_d2fridr(const gsl_function *f, double x, double h, 
		 double *err);


/***************************************************/
/* my_dfridr()                                     */
/* Compuate & evaluate the derivative of the func  */
/* f at the argument x. Based on the routine found */
/* in Numerical Recipes for C but modified for use */
/* with GSL vector, matrices and functional evalua-*/
/* tion.                                           */
/***************************************************/
double my_dfridr(const gsl_function *f, double x, double h, 
		 double *err)
{
    int i,j;
    double errt,fac,hh,ans;
    gsl_matrix *a;

    if(h ==0.0)
	{
	    fprintf(stderr,"\nh in subroutine: my_dfridr must be nonzero.\n");
	    exit(1);
	}

    a = gsl_matrix_alloc(_ntab,_ntab);
    hh = h;
    gsl_matrix_set(a,0,0, 
		   (GSL_FN_EVAL (f,x+hh) -  GSL_FN_EVAL (f,x-hh))/(2.0*hh) );
    *err = _big;
    for(i=1;i<_ntab;i++)
	{
	    hh /= _con;
	    gsl_matrix_set(a,0,i,
		   (GSL_FN_EVAL (f,x+hh) -  GSL_FN_EVAL (f,x-hh))/(2.0*hh) );
	    fac = _con2;
	   
	    for(j=1;j<=i;j++)
		{
		    gsl_matrix_set(a,j,i,
		       (gsl_matrix_get(a,j-1,i)*fac-gsl_matrix_get(a,j-1,i-1))/
				   (fac-1.0));
		    fac = _con2*fac;
		    errt=GSL_MAX(
		       fabs( gsl_matrix_get(a,j,i)-gsl_matrix_get(a,j-1,i) ),
                       fabs( gsl_matrix_get(a,j,i)-gsl_matrix_get(a,j-1,i-1)));

		    if (errt <= *err)
			{
			    *err = errt;
			    ans = gsl_matrix_get(a,j,i);
			}
		}
	    if (fabs(gsl_matrix_get(a,i,i)-gsl_matrix_get(a,i-1,i-1))
		>= _safe*(*err))
		break;
	}


    gsl_matrix_free(a);
    return ans;
}

/***************************************************/
/* my_d2fridr()                                    */
/* Compute & evaluates the second derivative of f  */
/* at the argument x using the central difference  */
/* formula approach of numerically compuating the  */
/* second derivative (see Cheney and Kincaid, Num- */
/* erical Mathematics and Computing, 5ed., p. 189  */
/***************************************************/

double my_d2fridr(const gsl_function *f, double x, double h, 
		 double *err)
{
    int i,j;
    double errt,fac,hh,ans;
    gsl_matrix *a;

    if(h ==0.0)
	{
	    fprintf(stderr,"\nh in subroutine: my_d2fridr must be nonzero.\n");
	    exit(1);
	}

    a = gsl_matrix_alloc(_ntab,_ntab);
    hh = h;
    gsl_matrix_set(a,0,0, 
		   (GSL_FN_EVAL (f,x+hh) - 2.0*GSL_FN_EVAL (f,x)
		    + GSL_FN_EVAL (f,x-hh))/pow(hh,2.0) );
    *err = _big;
    for(i=1;i<_ntab;i++)
	{
	    hh /= _con;
	    gsl_matrix_set(a,0,i,
			   (GSL_FN_EVAL (f,x+hh) - 2.0*GSL_FN_EVAL (f,x)
			    + GSL_FN_EVAL (f,x-hh))/pow(hh,2.0) );
	    fac = _con2;
	   
	    for(j=1;j<=i;j++)
		{
		    gsl_matrix_set(a,j,i,
		       (gsl_matrix_get(a,j-1,i)*fac-gsl_matrix_get(a,j-1,i-1))/
				   (fac-1.0));
		    fac = _con2*fac;
		    errt=GSL_MAX(
		       fabs( gsl_matrix_get(a,j,i)-gsl_matrix_get(a,j-1,i) ),
                       fabs( gsl_matrix_get(a,j,i)-gsl_matrix_get(a,j-1,i-1)));

		    if (errt <= *err)
			{
			    *err = errt;
			    ans = gsl_matrix_get(a,j,i);
			}
		}
	    if (fabs(gsl_matrix_get(a,i,i)-gsl_matrix_get(a,i-1,i-1))
		>= _safe*(*err))
		break;
	}


    gsl_matrix_free(a);
    return ans;
}